/* ENG is Not a Game (c) Thunor 2008 <thunorsif@hotmail.com>
www.thunor.org.uk

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */


/* Includes */
#include "ENG_engine.h"


/* Defines */
/*
#define DEBUG_ACTIONS
#define DEBUG_AXISVALUES
#define DEBUG_ACTIONSTATES
*/

/*#define ENG_ACTIONS_WITH_MODIFIERS_RELEASE_ON_BOTH*/
#define ENG_ACTIONS_WITH_MODIFIERS_RELEASE_ON_EITHER
#define ENG_KEY_REPEAT_DELAY 15
#define ENG_KEY_REPEAT_INTERVAL 5
#define ENG_JOYSTICK_DEAD_ZONE 50


/* Variables */
int eng_joystick_dead_zone = ENG_JOYSTICK_DEAD_ZONE;
int eng_key_repeat_delay = ENG_KEY_REPEAT_DELAY, eng_key_repeat_interval = ENG_KEY_REPEAT_INTERVAL;

ENG_Action eng_actions[ENG_ACTIONS];


/* Function prototypes */
void ENG_UnpressActions(ENG_Action *actions);


/***************************************************************************
 * Get Action                                                              *
 ***************************************************************************/
/* This abstracts the various input devices and their properties into actions.
   To each action can be assigned a single control and an additional modifier.
   Input devices can be mixed so a joystick button could have a keyboard modifier.
   Controls can be duplicated across actions.
   To each action can be assigned a unique delay and repeat interval.
   The delay or interval can be disabled by setting to zero (repeats with the minimum of delay).
   The delay or interval can be maximised by setting very high (repeats with the maximum of delay) e.g. 32767.
   When actions-with-modifiers are pressed, single actions using the same controls are released.
   Actions-with-modifiers can be released when both the action and modifier are
   released or when the action or modifier is released.

   How to call it :-
   pollall = FALSE to poll one SDL event to retrieve the control data in action.
   pollall = TRUE to get the first reported action press or release in actionid and
     again to get the next until actionid = ENG_UNDEFINED */
   
/* On entry: see above.
    On exit: actionid will contain an actionid on press
    		 actionid will contain an (actionid | 0x80) on release
    		   else actionid will contain ENG_UNDEFINED 
			 returns TRUE on SDL_QUIT else FALSE */

int ENG_GetAction(int *actionid, int pollall, ENG_Action *action) {
	static int axisvalues[ENG_MAX_JOYSTICKS][10 * 2];	/* I doubt there exists a joystick with more than 10 axes */
	static int lastactionid = ENG_UNDEFINED, init = TRUE;
	int count, count2, found = FALSE, axisplus, retval = FALSE, mx, my;
	SDL_Event event;

	if (init) {
		init = FALSE;
		/* Initialise the axisvalues the first time this function is executed */
		for (count = 0; count < ENG_MAX_JOYSTICKS; count++) {
			for (count2 = 0; count2 < 20; count2++) {
				axisvalues[count][count2] = 0;
			}
		}
	}
	
	*actionid = ENG_UNDEFINED;

	if (lastactionid == ENG_UNDEFINED || pollall == FALSE) {
		while (SDL_PollEvent(&event)) {
			action->device = action->index = action->id = action->state = ENG_UNDEFINED;
			switch (event.type) {
				case SDL_KEYUP:
				case SDL_KEYDOWN:
					action->device = ENG_DEVICE_KEYBOARD;
					action->id = event.key.keysym.sym;
					action->state = event.key.state;
					break;
				case SDL_JOYBUTTONUP:
				case SDL_JOYBUTTONDOWN:
					action->device = ENG_DEVICE_JOYSTICK;
					action->index = event.jbutton.which;
					action->id = event.jbutton.button;
					action->state = event.jbutton.state;
					break;
				case SDL_JOYAXISMOTION:
					/* Identify which end of the axis we are going to work with */
					axisplus = 0;	/* -32767 to 0 */
					if (event.jaxis.value > 0 || (event.jaxis.value == 0 && axisvalues[event.jaxis.which][event.jaxis.axis * 2 + 1] >= (32767 * eng_joystick_dead_zone / 100))) axisplus = 1;	/* 0 to +32767 */
					/* Does anything need to be reported as pressed or released? */
					if ((abs(event.jaxis.value) >= (32767 * eng_joystick_dead_zone / 100) && axisvalues[event.jaxis.which][event.jaxis.axis * 2 + axisplus] < (32767 * eng_joystick_dead_zone / 100)) ||
						(abs(event.jaxis.value) < (32767 * eng_joystick_dead_zone / 100) && axisvalues[event.jaxis.which][event.jaxis.axis * 2 + axisplus] >= (32767 * eng_joystick_dead_zone / 100))) {
						action->device = ENG_DEVICE_JOYSTICK;
						action->index = event.jaxis.which;
						action->id = event.jaxis.axis * 2 + axisplus + SDL_JoystickNumButtons(eng_joysticks[event.jaxis.which]);
						if (abs(event.jaxis.value) >= (32767 * eng_joystick_dead_zone / 100) && axisvalues[event.jaxis.which][event.jaxis.axis * 2 + axisplus] < (32767 * eng_joystick_dead_zone / 100)) {
							action->state = SDL_PRESSED;
						} else {
							action->state = SDL_RELEASED;
						}
						axisvalues[event.jaxis.which][event.jaxis.axis * 2 + axisplus] = abs(event.jaxis.value);
					}
					break;
				case SDL_MOUSEMOTION:
					if (eng_op_env.pointer_mode != ENG_POINTER_MODE_OFF) {
						/* eng_lyr_pointer is relative to the screen */
						eng_lyr_pointer->x = event.motion.x;
						eng_lyr_pointer->y = event.motion.y;
					}
					break;
				case SDL_MOUSEBUTTONUP:
				case SDL_MOUSEBUTTONDOWN:
					action->device = ENG_DEVICE_MOUSE;
					action->id = event.button.button;
					action->state = event.button.state;
					break;
				case SDL_QUIT:
					retval = TRUE;
					break;
				default:
					break;
			}

			if (!pollall) return retval;

			/* Update the state of all matching controls within the actions if found */
			if (action->device != ENG_UNDEFINED) {
				found = FALSE;
				for (count = 0; count < ENG_ACTIONS; count++) {
					if (eng_actions[count].device == action->device && eng_actions[count].index == action->index && eng_actions[count].id == action->id) {
						eng_actions[count].state = action->state;
						found = TRUE;
					}
				}
				if (found) break;	/* Stop polling events now as we've got something to work with */
			}						/* Actions can become pressed and released if we're not careful */
		}

		/* Press actions based on the current control states */
		for (count = 0; count < ENG_ACTIONS; count++) {
			if ((eng_actions[count].state == SDL_PRESSED && eng_actions[count].mod == ENG_UNDEFINED) ||
				(eng_actions[count].state == SDL_PRESSED && eng_actions[count].mod != ENG_UNDEFINED && eng_actions[eng_actions[count].mod].state == SDL_PRESSED)) {
				eng_actions[count].actionstate = SDL_PRESSED;
			}
		}
		/* Release actions based on the current control states */
		for (count = 0; count < ENG_ACTIONS; count++) {
			if ((eng_actions[count].mod == ENG_UNDEFINED && eng_actions[count].state == SDL_RELEASED) ||
				#if defined(ENG_ACTIONS_WITH_MODIFIERS_RELEASE_ON_BOTH)
					(eng_actions[count].mod != ENG_UNDEFINED && eng_actions[count].state == SDL_RELEASED && eng_actions[eng_actions[count].mod].state == SDL_RELEASED)) {	/* Release when the action and modifier are released */
				#elif defined(ENG_ACTIONS_WITH_MODIFIERS_RELEASE_ON_EITHER)
					(eng_actions[count].mod != ENG_UNDEFINED && (eng_actions[count].state == SDL_RELEASED || eng_actions[eng_actions[count].mod].state == SDL_RELEASED))) {	/* Release when the action or modifier is released */
				#endif
				eng_actions[count].actionstate = SDL_RELEASED;
			}
		}
		/* Are any actions-with-modifiers pressed? */
		for (count = 0; count < ENG_ACTIONS; count++) {
			if (eng_actions[count].actionstate == SDL_PRESSED && eng_actions[count].mod != ENG_UNDEFINED) {
				/* Release actions-without-modifiers using the same controls */
				for (count2 = 0; count2 < ENG_ACTIONS; count2++) {
					if (eng_actions[count2].actionstate == SDL_PRESSED && eng_actions[count2].mod == ENG_UNDEFINED &&
					((eng_actions[count2].device == eng_actions[count].device && eng_actions[count2].index == eng_actions[count].index && eng_actions[count2].id == eng_actions[count].id) ||
					(eng_actions[count2].device == eng_actions[eng_actions[count].mod].device && eng_actions[count2].index == eng_actions[eng_actions[count].mod].index && eng_actions[count2].id == eng_actions[eng_actions[count].mod].id))) {
						eng_actions[count2].actionstate = SDL_RELEASED;
					}
				}
			}
		}
		#ifdef DEBUG_ACTIONSTATES
			for (count = 0; count < ENG_ACTIONS; count++) {
				switch (count) {
					case ENG_ACTION_UP:
						printf("ENG_ACTION_UP                actionstate=%i state=%i modstate=%i oup=%p\n", eng_actions[count].actionstate, eng_actions[count].state, eng_actions[eng_actions[count].mod].state, eng_actions[count].oup);
						break;
					case ENG_ACTION_UP_RIGHT:
						printf("ENG_ACTION_UP_RIGHT          actionstate=%i state=%i modstate=%i oup=%p\n", eng_actions[count].actionstate, eng_actions[count].state, eng_actions[eng_actions[count].mod].state, eng_actions[count].oup);
						break;
					case ENG_ACTION_RIGHT:
						printf("ENG_ACTION_RIGHT             actionstate=%i state=%i modstate=%i oup=%p\n", eng_actions[count].actionstate, eng_actions[count].state, eng_actions[eng_actions[count].mod].state, eng_actions[count].oup);
						break;
					case ENG_ACTION_DOWN_RIGHT:
						printf("ENG_ACTION_DOWN_RIGHT        actionstate=%i state=%i modstate=%i oup=%p\n", eng_actions[count].actionstate, eng_actions[count].state, eng_actions[eng_actions[count].mod].state, eng_actions[count].oup);
						break;
					case ENG_ACTION_DOWN:
						printf("ENG_ACTION_DOWN              actionstate=%i state=%i modstate=%i oup=%p\n", eng_actions[count].actionstate, eng_actions[count].state, eng_actions[eng_actions[count].mod].state, eng_actions[count].oup);
						break;
					case ENG_ACTION_DOWN_LEFT:
						printf("ENG_ACTION_DOWN_LEFT         actionstate=%i state=%i modstate=%i oup=%p\n", eng_actions[count].actionstate, eng_actions[count].state, eng_actions[eng_actions[count].mod].state, eng_actions[count].oup);
						break;
					case ENG_ACTION_LEFT:
						printf("ENG_ACTION_LEFT              actionstate=%i state=%i modstate=%i oup=%p\n", eng_actions[count].actionstate, eng_actions[count].state, eng_actions[eng_actions[count].mod].state, eng_actions[count].oup);
						break;
					case ENG_ACTION_UP_LEFT:
						printf("ENG_ACTION_UP_LEFT           actionstate=%i state=%i modstate=%i oup=%p\n", eng_actions[count].actionstate, eng_actions[count].state, eng_actions[eng_actions[count].mod].state, eng_actions[count].oup);
						break;
					case ENG_ACTION_SELECT:
						printf("ENG_ACTION_SELECT            actionstate=%i state=%i modstate=%i oup=%p\n", eng_actions[count].actionstate, eng_actions[count].state, eng_actions[eng_actions[count].mod].state, eng_actions[count].oup);
						break;
					case ENG_ACTION_SHUTDOWN:
						printf("ENG_ACTION_SHUTDOWN          actionstate=%i state=%i modstate=%i oup=%p\n", eng_actions[count].actionstate, eng_actions[count].state, eng_actions[eng_actions[count].mod].state, eng_actions[count].oup);
						break;
					case ENG_ACTION_TOGGLE_FULLSCREEN:
						printf("ENG_ACTION_TOGGLE_FULLSCREEN actionstate=%i state=%i modstate=%i oup=%p\n", eng_actions[count].actionstate, eng_actions[count].state, eng_actions[eng_actions[count].mod].state, eng_actions[count].oup);
						break;
					case ENG_ACTION_MODIFIER1:
						printf("ENG_ACTION_MODIFIER1         actionstate=%i state=%i modstate=%i oup=%p\n", eng_actions[count].actionstate, eng_actions[count].state, eng_actions[eng_actions[count].mod].state, eng_actions[count].oup);
						break;
					case ENG_ACTION_MODIFIER2:
						printf("ENG_ACTION_MODIFIER2         actionstate=%i state=%i modstate=%i oup=%p\n", eng_actions[count].actionstate, eng_actions[count].state, eng_actions[eng_actions[count].mod].state, eng_actions[count].oup);
						break;
				}
			}
		#endif
		/* Store object information for newly pressed actions */
		for (count = 0; count < ENG_ACTIONS; count++) {
			if (eng_actions[count].actionstate == SDL_PRESSED && eng_actions[count].oup == NULL) {
				/* Get the current pointer coordinates */
				ENG_GetCanvasOffset(eng_lyr_pointer, &mx, &my);
				eng_actions[count].mx = mx;
				eng_actions[count].my = my;
				eng_actions[count].oup = ENG_GetObjectUnderObject(eng_lyr_pointer, ENG_HIT_POINT_TOP_LEFT);
				eng_actions[count].cyclesactive = 0;
			}
		}
		
		/** Now it is possible to report action releases and presses to the caller */
		/** RELEASES */
		/* If an action was recently released then report it NOW */
		for (count = 0; count < ENG_ACTIONS; count++) {
			if (eng_actions[count].actionstate == SDL_RELEASED && eng_actions[count].oup != NULL) {
				*actionid = count | 0x80;
				eng_actions[count].oup = NULL;	/* Reset */
				return retval;
			}
		}
	}

	#ifdef DEBUG_AXISVALUES
		for (count = 0; count < ENG_MAX_JOYSTICKS; count++) {
			for (count2 = 0; count2 < 20; count2++) {
				printf("%i, ", axisvalues[count][count2]);
			}
			printf("\n");
		}
	#endif

	do {
		/** PRESSES */
		found = FALSE;
		/* If an action was recently pressed then report it */
		if (eng_actions[++lastactionid].actionstate == SDL_PRESSED) {
			found = TRUE;
			*actionid = lastactionid;
		}
		
		/* Because this function is called every main loop cycle, if an action is being pressed
		   then it would trigger an action every main loop cycle. cyclesactive, delay and interval
		   are used to decide the frequency of individual repeats */
		if (found) {
			if (eng_actions[*actionid].cyclesactive == 0) {
				/* Allow this action */
			} else if (eng_actions[*actionid].cyclesactive == eng_actions[*actionid].delay) {
				/* Allow this action */
			} else if (eng_actions[*actionid].cyclesactive > eng_actions[*actionid].delay) {
				if (eng_actions[*actionid].interval == 0) {
					/* Allow this action */
				} else if ((eng_actions[*actionid].cyclesactive - eng_actions[*actionid].delay) % eng_actions[*actionid].interval == 0) {
					/* Allow this action */
				} else {
					found = FALSE;
				}
			} else {
				found = FALSE;
			}
			eng_actions[*actionid].cyclesactive++;
			if (!found) *actionid = ENG_UNDEFINED;
		} else {
			*actionid = ENG_UNDEFINED;
		}
	} while (lastactionid < ENG_ACTIONS && !found);
	
	if (lastactionid >= ENG_ACTIONS) lastactionid = ENG_UNDEFINED;
		
	return retval;
}

/***************************************************************************
 * Unpress Actions                                                         *
 ***************************************************************************/
/* This isn't used anymore */

void ENG_UnpressActions(ENG_Action *actions) {
	int count;
	
	for (count = 0; count < ENG_ACTIONS; count++) {
		actions[count].state = actions[count].actionstate = SDL_RELEASED;
		actions[count].oup = NULL;
	}
}

/***************************************************************************
 * Show Action                                                             *
 ***************************************************************************/

void ENG_ShowAction(int *actionid) {
	char temp[256];

	switch (*actionid & 0x7f) {
		case ENG_ACTION_UP:
			strcpy(temp, "ENG_ACTION_UP");
			break;
		case ENG_ACTION_UP_RIGHT:
			strcpy(temp, "ENG_ACTION_UP_RIGHT");
			break;
		case ENG_ACTION_RIGHT:
			strcpy(temp, "ENG_ACTION_RIGHT");
			break;
		case ENG_ACTION_DOWN_RIGHT:
			strcpy(temp, "ENG_ACTION_DOWN_RIGHT");
			break;
		case ENG_ACTION_DOWN:
			strcpy(temp, "ENG_ACTION_DOWN");
			break;
		case ENG_ACTION_DOWN_LEFT:
			strcpy(temp, "ENG_ACTION_DOWN_LEFT");
			break;
		case ENG_ACTION_LEFT:
			strcpy(temp, "ENG_ACTION_LEFT");
			break;
		case ENG_ACTION_UP_LEFT:
			strcpy(temp, "ENG_ACTION_UP_LEFT");
			break;
		case ENG_ACTION_SELECT:
			strcpy(temp, "ENG_ACTION_SELECT");
			break;
		case ENG_ACTION_SHUTDOWN:
			strcpy(temp, "ENG_ACTION_SHUTDOWN");
			break;
		case ENG_ACTION_TOGGLE_FULLSCREEN:
			strcpy(temp, "ENG_ACTION_TOGGLE_FULLSCREEN");
			break;
		case ENG_ACTION_MODIFIER1:
			strcpy(temp, "ENG_ACTION_MODIFIER1");
			break;
		case ENG_ACTION_MODIFIER2:
			strcpy(temp, "ENG_ACTION_MODIFIER2");
			break;
		case ENG_UNDEFINED:
			strcpy(temp, "ENG_UNDEFINED");
			break;
		default:
			sprintf(temp, "Unknown");
			break;
	}

	if (*actionid & 0x80) {
		printf("%s: %s (%i) release, mx=%i my=%i cyclesactive=%i SDL_GetTicks=%i\n", __func__, temp, *actionid, eng_actions[*actionid & 0x7f].mx, eng_actions[*actionid & 0x7f].my, eng_actions[*actionid & 0x7f].cyclesactive, SDL_GetTicks());
	} else {
		printf("%s: %s (%i) press, mx=%i my=%i cyclesactive=%i SDL_GetTicks=%i\n", __func__, temp, *actionid, eng_actions[*actionid].mx, eng_actions[*actionid].my, eng_actions[*actionid].cyclesactive, SDL_GetTicks());
	}

}

/***************************************************************************
 * Initialise Actions                                                      *
 ***************************************************************************/

void ENG_InitActions(void) {
	int count;

	/* Set some defaults */
	for (count = 0; count < ENG_ACTIONS; count++) {
		#if defined(PLATFORM_PC)
			eng_actions[count].device = ENG_DEVICE_KEYBOARD;
			eng_actions[count].index = ENG_UNDEFINED;
		#elif defined(PLATFORM_ZAURUS)
			eng_actions[count].device = ENG_DEVICE_KEYBOARD;
			eng_actions[count].index = ENG_UNDEFINED;
		#elif defined(PLATFORM_GP2X)
			eng_actions[count].device = ENG_DEVICE_JOYSTICK;
			eng_actions[count].index = 0;
		#endif
		eng_actions[count].id = ENG_UNDEFINED;
		eng_actions[count].mod = ENG_UNDEFINED;
		eng_actions[count].state = SDL_RELEASED;
		eng_actions[count].actionstate = SDL_RELEASED;
		eng_actions[count].mx = ENG_UNDEFINED;
		eng_actions[count].my = ENG_UNDEFINED;
		eng_actions[count].oup = NULL;
		eng_actions[count].cyclesactive = 0;
		eng_actions[count].delay = eng_key_repeat_delay;
		eng_actions[count].interval = eng_key_repeat_interval;
	}
	
	/* Now tune to a platform */
	#if defined(PLATFORM_PC)
		eng_actions[ENG_ACTION_UP].id = SDLK_UP;
		eng_actions[ENG_ACTION_UP].delay = 0;
		eng_actions[ENG_ACTION_UP].interval = 0;
		/*eng_actions[ENG_ACTION_UP_RIGHT].id = SDLK_UP;
		eng_actions[ENG_ACTION_UP_RIGHT].mod = ENG_ACTION_RIGHT;
		eng_actions[ENG_ACTION_UP_RIGHT].delay = 0;
		eng_actions[ENG_ACTION_UP_RIGHT].interval = 0; */
		eng_actions[ENG_ACTION_RIGHT].id = SDLK_RIGHT;
		eng_actions[ENG_ACTION_RIGHT].delay = 0;
		eng_actions[ENG_ACTION_RIGHT].interval = 0;
		/*eng_actions[ENG_ACTION_DOWN_RIGHT].id = SDLK_DOWN;
		eng_actions[ENG_ACTION_DOWN_RIGHT].mod = ENG_ACTION_RIGHT;
		eng_actions[ENG_ACTION_DOWN_RIGHT].delay = 0;
		eng_actions[ENG_ACTION_DOWN_RIGHT].interval = 0; */
		eng_actions[ENG_ACTION_DOWN].id = SDLK_DOWN;
		eng_actions[ENG_ACTION_DOWN].delay = 0;
		eng_actions[ENG_ACTION_DOWN].interval = 0;
		/*eng_actions[ENG_ACTION_DOWN_LEFT].id = SDLK_DOWN;
		eng_actions[ENG_ACTION_DOWN_LEFT].mod = ENG_ACTION_LEFT;
		eng_actions[ENG_ACTION_DOWN_LEFT].delay = 0;
		eng_actions[ENG_ACTION_DOWN_LEFT].interval = 0; */
		eng_actions[ENG_ACTION_LEFT].id = SDLK_LEFT;
		eng_actions[ENG_ACTION_LEFT].delay = 0;
		eng_actions[ENG_ACTION_LEFT].interval = 0;
		/*eng_actions[ENG_ACTION_UP_LEFT].id = SDLK_UP;
		eng_actions[ENG_ACTION_UP_LEFT].mod = ENG_ACTION_LEFT;
		eng_actions[ENG_ACTION_UP_LEFT].delay = 0;
		eng_actions[ENG_ACTION_UP_LEFT].interval = 0; */
		eng_actions[ENG_ACTION_SELECT].device = ENG_DEVICE_MOUSE;
		eng_actions[ENG_ACTION_SELECT].id = SDL_BUTTON_LEFT;
		//eng_actions[ENG_ACTION_SELECT].delay = 0;	/* temp temp */
		//eng_actions[ENG_ACTION_SELECT].interval = 0;	/* temp temp */
		//eng_actions[ENG_ACTION_SELECT].id = SDLK_RETURN;
		eng_actions[ENG_ACTION_SHUTDOWN].id = SDLK_q;
		eng_actions[ENG_ACTION_TOGGLE_FULLSCREEN].id = SDLK_RETURN;
		eng_actions[ENG_ACTION_TOGGLE_FULLSCREEN].mod = ENG_ACTION_MODIFIER1;
		eng_actions[ENG_ACTION_MODIFIER1].id = SDLK_LALT;
		eng_actions[ENG_ACTION_MODIFIER2].id = SDLK_RSHIFT;
	#elif defined(PLATFORM_ZAURUS)
		eng_actions[ENG_ACTION_UP].id = ZAURUS_UP;
		eng_actions[ENG_ACTION_UP].delay = 0;
		eng_actions[ENG_ACTION_UP].interval = 0;
		eng_actions[ENG_ACTION_UP_RIGHT].id = ZAURUS_UP;
		eng_actions[ENG_ACTION_UP_RIGHT].mod = ENG_ACTION_RIGHT;
		eng_actions[ENG_ACTION_UP_RIGHT].delay = 0;
		eng_actions[ENG_ACTION_UP_RIGHT].interval = 0;
		eng_actions[ENG_ACTION_RIGHT].id = ZAURUS_RIGHT;
		eng_actions[ENG_ACTION_RIGHT].delay = 0;
		eng_actions[ENG_ACTION_RIGHT].interval = 0;
		eng_actions[ENG_ACTION_DOWN_RIGHT].id = ZAURUS_DOWN;
		eng_actions[ENG_ACTION_DOWN_RIGHT].mod = ENG_ACTION_RIGHT;
		eng_actions[ENG_ACTION_DOWN_RIGHT].delay = 0;
		eng_actions[ENG_ACTION_DOWN_RIGHT].interval = 0;
		eng_actions[ENG_ACTION_DOWN].id = ZAURUS_DOWN;
		eng_actions[ENG_ACTION_DOWN].delay = 0;
		eng_actions[ENG_ACTION_DOWN].interval = 0;
		eng_actions[ENG_ACTION_DOWN_LEFT].id = ZAURUS_DOWN;
		eng_actions[ENG_ACTION_DOWN_LEFT].mod = ENG_ACTION_LEFT;
		eng_actions[ENG_ACTION_DOWN_LEFT].delay = 0;
		eng_actions[ENG_ACTION_DOWN_LEFT].interval = 0;
		eng_actions[ENG_ACTION_LEFT].id = ZAURUS_LEFT;
		eng_actions[ENG_ACTION_LEFT].delay = 0;
		eng_actions[ENG_ACTION_LEFT].interval = 0;
		eng_actions[ENG_ACTION_UP_LEFT].id = ZAURUS_UP;
		eng_actions[ENG_ACTION_UP_LEFT].mod = ENG_ACTION_LEFT;
		eng_actions[ENG_ACTION_UP_LEFT].delay = 0;
		eng_actions[ENG_ACTION_UP_LEFT].interval = 0;
		eng_actions[ENG_ACTION_SELECT].device = ENG_DEVICE_MOUSE;
		eng_actions[ENG_ACTION_SELECT].id = SDL_BUTTON_LEFT;
		//eng_actions[ENG_ACTION_SELECT].delay = 0;	/* temp temp */
		//eng_actions[ENG_ACTION_SELECT].interval = 0;	/* temp temp */
		//eng_actions[ENG_ACTION_SELECT].id = ZAURUS_SPACE;
		eng_actions[ENG_ACTION_SHUTDOWN].id = ZAURUS_HOME;
		eng_actions[ENG_ACTION_TOGGLE_FULLSCREEN].id = ZAURUS_ADDRESS;
		eng_actions[ENG_ACTION_MODIFIER1].id = ZAURUS_MENU;
		eng_actions[ENG_ACTION_MODIFIER2].id = ZAURUS_EMAIL;
	#elif defined(PLATFORM_GP2X)
		eng_actions[ENG_ACTION_UP].id = GP2X_JOY_N;
		eng_actions[ENG_ACTION_UP].delay = 0;
		eng_actions[ENG_ACTION_UP].interval = 0;
		eng_actions[ENG_ACTION_UP_RIGHT].id = GP2X_JOY_NE;
		eng_actions[ENG_ACTION_UP_RIGHT].delay = 0;
		eng_actions[ENG_ACTION_UP_RIGHT].interval = 0;
		eng_actions[ENG_ACTION_RIGHT].id = GP2X_JOY_E;
		eng_actions[ENG_ACTION_RIGHT].delay = 0;
		eng_actions[ENG_ACTION_RIGHT].interval = 0;
		eng_actions[ENG_ACTION_DOWN_RIGHT].id = GP2X_JOY_SE;
		eng_actions[ENG_ACTION_DOWN_RIGHT].delay = 0;
		eng_actions[ENG_ACTION_DOWN_RIGHT].interval = 0;
		eng_actions[ENG_ACTION_DOWN].id = GP2X_JOY_S;
		eng_actions[ENG_ACTION_DOWN].delay = 0;
		eng_actions[ENG_ACTION_DOWN].interval = 0;
		eng_actions[ENG_ACTION_DOWN_LEFT].id = GP2X_JOY_SW;
		eng_actions[ENG_ACTION_DOWN_LEFT].delay = 0;
		eng_actions[ENG_ACTION_DOWN_LEFT].interval = 0;
		eng_actions[ENG_ACTION_LEFT].id = GP2X_JOY_W;
		eng_actions[ENG_ACTION_LEFT].delay = 0;
		eng_actions[ENG_ACTION_LEFT].interval = 0;
		eng_actions[ENG_ACTION_UP_LEFT].id = GP2X_JOY_NW;
		eng_actions[ENG_ACTION_UP_LEFT].delay = 0;
		eng_actions[ENG_ACTION_UP_LEFT].interval = 0;
		eng_actions[ENG_ACTION_SELECT].id =  GP2X_BTN_B;
		//eng_actions[ENG_ACTION_SELECT].delay = 0;	/* temp temp */
		//eng_actions[ENG_ACTION_SELECT].interval = 0;	/* temp temp */
		eng_actions[ENG_ACTION_SHUTDOWN].id = GP2X_LTRIG;
		eng_actions[ENG_ACTION_SHUTDOWN].mod = ENG_ACTION_MODIFIER1;
		//eng_actions[ENG_ACTION_TOGGLE_FULLSCREEN].id = ENG_UNDEFINED;
		eng_actions[ENG_ACTION_TOGGLE_FULLSCREEN].id = GP2X_BTN_Y;
		eng_actions[ENG_ACTION_MODIFIER1].id = GP2X_RTRIG;
		eng_actions[ENG_ACTION_MODIFIER2].id = GP2X_BTN_A;
	#endif

}

