/* ENG is Not a Game (c) Thunor 2008 <thunorsif@hotmail.com>
www.thunor.org.uk

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */


/* Includes */
#include "ENG_engine.h"


/* Defines */
/*
#define DEBUG_CONFIG
#define DEBUG_CONFIG_FILE
*/

#define ENG_MAX_CONFIG_OPTIONS 1000


/* Variables */
typedef struct engconfigoption {
	char option[ENG_CONFIG_OPTION_LENGTH];
} ENG_ConfigOption;

ENG_ConfigOption *eng_config_options[ENG_MAX_CONFIG_OPTIONS];
int eng_config_options_count;
extern int eng_inited;


/* Function prototypes */
int ENG_CreateConfigOption(ENG_ConfigOption *option);
int ENG_FreeConfigOption(ENG_ConfigOption *option);


/***************************************************************************
 * Initialise Config                                                       *
 ***************************************************************************/

void ENG_InitConfig(void) {
	eng_config_options_count = 0;
}

/***************************************************************************
 * Load Config                                                             *
 ***************************************************************************/
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_LoadConfig(void) {
	int retval = ENG_ERROR_NONE, count, linecount = 0;
	char line[256];
	FILE *fp;
	ENG_ConfigOption option;
	
	#ifdef DEBUG_CONFIG
		printf("%s: Loading\n", __func__);
	#endif
	/* Load the config options from the config file */
	printf("%s: Reading from %s\n", __func__, eng_op_env.config_file);
	/* Open config file */
	if ((fp = fopen(eng_op_env.config_file, "r")) == NULL) {
		sprintf(eng_last_error, "%s: %s not found", __func__, eng_op_env.config_file);
		retval = ENG_ERROR_CANNOT_OPEN_CONFIG_FILE;
	} else {
		/* Read data until no more lines */
		while (fgets(line, 256, fp) != NULL) {	/* Reads the string + one or more [CR]LFs */
			linecount++;
			/* Chop off the [CR]LFs as we'll put them back when saving */
			for (count = strlen(line) - 1; count >= 0; count--) {
				if (line[count] == 13 || line[count] == 10) line[count] = 0;
			}
			#ifdef DEBUG_CONFIG_FILE
				printf("line %03i: %s\n", linecount, line);
			#endif
			/* Validate line length */
			if (strlen(line) >= ENG_CONFIG_OPTION_LENGTH) {
				sprintf(eng_last_error, "%s: Configuration option in %s at line %i is too long (max %i)", __func__, eng_op_env.config_file, linecount, ENG_CONFIG_OPTION_LENGTH - 1);
				retval = ENG_ERROR_CONFIG_FILE_SYNTAX_ERROR;
				break;
			}
			/* Create config option */
			strcpy(option.option, line);
			retval = ENG_CreateConfigOption(&option);
			if (retval) break;
		}
		fclose(fp);
	}

	return retval;
}

/***************************************************************************
 * Create Config Option                                                    *
 ***************************************************************************/
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_CreateConfigOption(ENG_ConfigOption *option) {
	int retval = ENG_ERROR_NONE;
	ENG_ConfigOption *newoption;

	if (eng_config_options_count >= ENG_MAX_CONFIG_OPTIONS) {
		sprintf(eng_last_error, "%s: Too many configuration options (max %i)", __func__, ENG_MAX_CONFIG_OPTIONS);
		retval = ENG_ERROR_TOO_MANY_CONFIG_OPTIONS;
	} else {
		newoption = (ENG_ConfigOption*) malloc(sizeof(ENG_ConfigOption));
		*newoption = *option;
		#ifdef DEBUG_CONFIG
			printf("%s: Creating %p option=%s\n", __func__, newoption, newoption->option);
		#endif
		eng_config_options[eng_config_options_count] = newoption;
		eng_config_options_count++;
	}

	return retval;
}

/***************************************************************************
 * Save Config                                                             *
 ***************************************************************************/
/* This can be called after ENG_Init and before ENG_Start so it's possible
   to load the rcfile, set something, save it and quit without starting the
   engine if that's what the user wants */

/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_SaveConfig(void) {
	int retval = ENG_ERROR_NONE, count;
	FILE *fp;

	#ifdef DEBUG_CONFIG
		printf("%s: Saving\n", __func__);
	#endif
	/* Save all config options to the config file */
	printf("%s: Writing to %s\n", __func__, eng_op_env.config_file);
	/* Open config file */
	if ((fp = fopen(eng_op_env.config_file, "w")) == NULL) {
		sprintf(eng_last_error, "%s: Cannot write to %s", __func__, eng_op_env.config_file);
		retval = ENG_ERROR_CANNOT_WRITE_TO_CONFIG_FILE;
	} else {
		/* Simply dump the array contents */
		for (count = 0; count < eng_config_options_count; count++) {
			fprintf(fp, "%s\n", eng_config_options[count]->option);
		}
		fclose(fp);
	}
	
	return retval;
}

/***************************************************************************
 * Get Config Option                                                       *
 ***************************************************************************/
/* This retrieves the value part of the configuration option identified by
   name. If name is not found then an empty string is returned */
   
/* On entry: name points to a string that identifies the config option
			 value points to a string to place the value or ""
    On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_GetConfigOption(char *name, char *value) {
	int retval = ENG_ERROR_NONE, count, found = FALSE;
	char localname[ENG_CONFIG_OPTION_NAME_LENGTH + 1];
	
	if (eng_inited == FALSE) {
		sprintf(eng_last_error, "%s: ENG has not been initialised yet", __func__);
		retval = ENG_ERROR_ENG_NOT_YET_INITIALISED;
	} else if (strlen(name) >= ENG_CONFIG_OPTION_NAME_LENGTH) {
		sprintf(eng_last_error, "%s: name is too long (max %i)", __func__, ENG_CONFIG_OPTION_NAME_LENGTH - 1);
		retval = ENG_ERROR_INVALID_CONFIG_OPTION;
	} else {
		/* Copy name locally and append it with an '=' */
		strcpy(localname, name);
		strcat(localname, "=");
		/* Search for localname */
		for (count = 0; count < eng_config_options_count; count++) {
			/* Have we found a matching name? */
			if (!strncmp(eng_config_options[count]->option, localname, strlen(localname))) {
				/* Is the value part short enough? */
				if (strlen(&eng_config_options[count]->option[strlen(localname)]) >= ENG_CONFIG_OPTION_VALUE_LENGTH) {
					sprintf(eng_last_error, "%s: value is too long (max %i)", __func__, ENG_CONFIG_OPTION_VALUE_LENGTH - 1);
					retval = ENG_ERROR_INVALID_CONFIG_OPTION;
					break;
				} else {
					/* Copy value part to value and return */
					strcpy(value, &eng_config_options[count]->option[strlen(localname)]);
					found = TRUE;
					break;
				}
			}
		}
		/* If localname was not found then return an empty string */
		if (!retval && !found) strcpy(value, "");
	}
	
	return retval;
}

/***************************************************************************
 * Set Config Option                                                       *
 ***************************************************************************/
/* This sets the value part of the configuration option identified by
   name. If name is not found then a new config option is created */
   
/* On entry: name points to a string that identifies the config option
			 value points to a string that holds the value to set
    On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_SetConfigOption(char *name, char *value) {
	int retval = ENG_ERROR_NONE, count, found = FALSE;
	char localname[ENG_CONFIG_OPTION_NAME_LENGTH + 1];
	ENG_ConfigOption option;
	
	if (eng_inited == FALSE) {
		sprintf(eng_last_error, "%s: ENG has not been initialised yet", __func__);
		retval = ENG_ERROR_ENG_NOT_YET_INITIALISED;
	} else if (strlen(name) >= ENG_CONFIG_OPTION_NAME_LENGTH) {
		sprintf(eng_last_error, "%s: name is too long (max %i)", __func__, ENG_CONFIG_OPTION_NAME_LENGTH - 1);
		retval = ENG_ERROR_INVALID_CONFIG_OPTION;
	} else if (strlen(value) >= ENG_CONFIG_OPTION_VALUE_LENGTH) {
		sprintf(eng_last_error, "%s: value is too long (max %i)", __func__, ENG_CONFIG_OPTION_VALUE_LENGTH - 1);
		retval = ENG_ERROR_INVALID_CONFIG_OPTION;
	} else {
		/* Copy name locally and append it with an '=' */
		strcpy(localname, name);
		strcat(localname, "=");
		/* Search for localname */
		for (count = 0; count < eng_config_options_count; count++) {
			/* Have we found a matching name? */
			if (!strncmp(eng_config_options[count]->option, localname, strlen(localname))) {
				/* Replace value part with new value  */
				strcpy(&eng_config_options[count]->option[strlen(localname)], value);
				found = TRUE;
				break;
			}
		}
		/* If localname was not found then create the config option */
		if (!found) {
			strcpy(option.option, localname);
			strcat(option.option, value);
			retval = ENG_CreateConfigOption(&option);
		}
	}
	
	return retval;
}

/***************************************************************************
 * Free Config Option                                                      *
 ***************************************************************************/
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */
	
int ENG_FreeConfigOption(ENG_ConfigOption *option) {
	int retval = ENG_ERROR_NONE, count, count2;

	if (eng_config_options_count <= 0) {
		sprintf(eng_last_error, "%s: There is no option to free", __func__);
		retval = ENG_ERROR_INVALID_CONFIG_OPTION;
	} else if (option == NULL || sizeof(*option) != sizeof(ENG_ConfigOption)) {
		sprintf(eng_last_error, "%s: option is invalid", __func__);
		retval = ENG_ERROR_INVALID_CONFIG_OPTION;
	} else {
		#ifdef DEBUG_CONFIG
			printf("%s: Freeing %p option=%s\n", __func__, option, option->option);
		#endif
		/* If anything is above the current option then move it down the array */
		for (count = 0; count < eng_config_options_count - 1; count++) {
			if (eng_config_options[count] == option) {
				for (count2 = count; count2 < eng_config_options_count - 1; count2++) {
					eng_config_options[count2] = eng_config_options[count2 + 1];
				}
				break;				
			}
		}
		/* Free the memory and update the array element count now */
		free(option);
		eng_config_options_count--;
	}

	return retval;
}

/***************************************************************************
 * Free All Config Options                                                 *
 ***************************************************************************/
/* This mops-up any remaining config options before shutting down */

void ENG_FreeAllConfigOptions(void) {
	int count;
	
	if (eng_config_options_count > 0) {
		printf("%s: Freeing\n", __func__);
		/* Free all remaining config options */
		for (count = eng_config_options_count - 1; count >= 0; count--) {
			ENG_FreeConfigOption(eng_config_options[count]);
		}
	}
}





