/* ENG is Not a Game (c) Thunor 2008 <thunorsif@hotmail.com>
www.thunor.org.uk

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */


/* Includes */
#include "ENG_engine.h"


/* Defines */
/*
#define DEBUG_EVENTS
*/


/* Variables */


/* Function prototypes */
void ENG_ManagePointerInput(int pxu, int pyu);
void ENG_EventProcessor(ENG_Event *event);


/***************************************************************************
 * Initialise Events                                                       *
 ***************************************************************************/

void ENG_InitEvents(void) {
}


/***************************************************************************
 * Generate Events                                                         *
 ***************************************************************************/
/* This is where the user gets to interact with the engine.
   
   This is how it currently works and it's uncomplicated :-
   
   ENG_POINTER_MODE_OFF
   --------------------
   The [simulated] pointer is not operational.
   Cursor events (up, down etc.) are sent to the user's event processor.
   The object attributed to all events is ENG_LYR_ROOT.
   All pointer code in this mode is disabled although the pointer visibility
   can still be manipulated as can any physical system pointer hardware.
   If you want to manage your own pointer and object highlighting then use
   this mode.
   
   ENG_POINTER_MODE_PHYSICAL
   -------------------------
   The [simulated] pointer is operational but not via the cursor events.
   Cursor events (up, down etc.) are sent to the user's event processor.
   The user's event processor will receive any of 6 pointer related events:
   press, click, release, drag, over and out, and all of them are directly
   related to the objects they operate upon.
   Press works thus: If the user presses an enabled object then one press
   event is generated.
   Click works thus: If the user presses an enabled object and releases
   over the same object then one click event is generated.
   Release works thus: If the user presses an enabled object and releases
   over a different object then one release event is generated.
   Drag works thus: If the user presses an enabled object and moves the
   pointer then a drag event is generated.
   Whilst an object is being pressed and the pointer remains over the same
   object, a select event is generated and will continue to do so as defined
   by the action's repeat delay and interval values.
   An over event is generated (once) when entering an enabled object's area
   and an out event is generated (once) when leaving the object for another.
   If you want to use any physical mouse pointer or stylus and have other uses
   for the cursor events then use this mode.

   ENG_POINTER_MODE_SIMULATED
   --------------------------
   The [simulated] pointer is operational via the cursor events.
   Cursor events (up, down etc.) are NOT sent to the user's event processor.
   Other than that this mode is identical to ENG_POINTER_MODE_PHYSICAL.
   If you don't have any physical mouse pointer or stylus and would like to
   use a simulated pointer controllable via the cursor events then use this mode.
   
   ENG_POINTER_MODE_NAVIGATION
   ---------------------------
   Note yet implemented.

   Non Pointer Related Events
   --------------------------
   System generated events such as ENG_EVENT_QUITTING and ENG_EVENT_POST_RENDER
   are attributed to ENG_LYR_ROOT.
   Unless the pointer mode is off, all other events will be attributed to the
   object currently under the pointer but are not directly related to the events.
   
   Use ENG_ShowEvent(event) within your event processor to gain a more complete
   understanding of the events. */

void ENG_GenerateEvents(void) {
	static int last_pointer_x = 0, last_pointer_y = 0;
	static int press_point_x = 0, press_point_y = 0;
	int actionid, mx, my, px, py, pxu = 0, pyu = 0;
	ENG_Action action;
	ENG_Event event;
	ENG_Object *new_oup, *pob;

	#ifdef ENG_EVENT_SEND_CYCLE
		/* Alert the user to the next cycle */
		event.object = eng_lyr_root; event.klasse = ENG_EVENT_CYCLE; ENG_SendEvent(&event);
	#endif

	/* Process all available actions */
	do {
		if ((ENG_GetAction(&actionid, TRUE, &action))) {
			ENG_Quit();
			break;
		} else {
			if (eng_op_env.pointer_mode == ENG_POINTER_MODE_OFF) {
				eng_oup = eng_lyr_root;
			} else {
				/* Get or set the object currently under the pointer. If an object is currently
				   being pressed and it's enabled and draggable then it is attached to the pointer */
				if (eng_pressed_object != NULL && ENG_GetEnability(eng_pressed_object) && eng_pressed_object->draggable) {
					/* Force it as eng_oup to prevent dragging-off through latency */
					eng_oup = eng_pressed_object;
					ENG_GetCanvasOffset(eng_oup->pob, &px, &py);
					/* Attach the object to the pointer */
					ENG_GetCanvasOffset(eng_lyr_pointer, &mx, &my);
					eng_oup->x = mx - px - press_point_x;
					eng_oup->y = my - py - press_point_y;
					/* Is dragging restricted in any way? */
					if (eng_oup->draggable == ENG_DRAG_RESTRICT_TO_PARENT) {
						pob = eng_oup->pob;
						if (eng_oup->x < 0) {
							eng_oup->x = 0;
						} else if (eng_oup->x + eng_oup->w > pob->w) {
							eng_oup->x = pob->w - eng_oup->w;
						}
						if (eng_oup->y < 0) {
							eng_oup->y = 0;
						} else if (eng_oup->y + eng_oup->h > pob->h) {
							eng_oup->y = pob->h - eng_oup->h;
						}
					}
				} else {
					new_oup = ENG_GetObjectUnderObject(eng_lyr_pointer, ENG_HIT_POINT_TOP_LEFT);
					/* eng_oup could be NULL here because it may have been a freed object but
					   new_oup will most definitely contain at least a system object */
					if (eng_oup != new_oup) {
						/* Send one out event if the object is not NULL and enabled */
						if (eng_oup != NULL && ENG_GetEnability(eng_oup)) {
							event.object = eng_oup; event.klasse = ENG_EVENT_OUT; ENG_SendEvent(&event);
						}
						eng_oup = new_oup;
						/* Send one over event if the object is enabled */
						if (ENG_GetEnability(eng_oup)) {
							event.object = eng_oup; event.klasse = ENG_EVENT_OVER; ENG_SendEvent(&event);
						}
					}
					ENG_GetCanvasOffset(eng_oup->pob, &px, &py);
				}
				/* Detect dragging of the pointer across enabled objects */
				if (eng_pressed_object != NULL && eng_pressed_object == eng_oup && ENG_GetEnability(eng_oup)) {
					if (last_pointer_x != eng_lyr_pointer->x || last_pointer_y != eng_lyr_pointer->y) {
						event.object = eng_oup; event.klasse = ENG_EVENT_DRAG; ENG_SendEvent(&event);
					}
				}
				/* Record current pointer offsets for later comparison */
				last_pointer_x = eng_lyr_pointer->x;
				last_pointer_y = eng_lyr_pointer->y;
			}

			/* Process any reported action */
			if (actionid != ENG_UNDEFINED) {

				#ifdef DEBUG_ACTIONS
					ENG_ShowAction(&actionid);
				#endif
				
				/* Convert actions to events and send them to the event processor. Note that apart from
				   the select event, the object is not directly related but may be useful to the user to
				   detect the object currently underneath the pointer */
				switch (actionid) {
					case ENG_ACTION_UP:
						if (eng_op_env.pointer_mode == ENG_POINTER_MODE_OFF || eng_op_env.pointer_mode == ENG_POINTER_MODE_PHYSICAL) {
							event.object = eng_oup; event.klasse = ENG_EVENT_UP; ENG_SendEvent(&event);
						} else {
							pyu = -1;
						}
						break;
					case ENG_ACTION_UP_RIGHT:
						if (eng_op_env.pointer_mode == ENG_POINTER_MODE_OFF || eng_op_env.pointer_mode == ENG_POINTER_MODE_PHYSICAL) {
							event.object = eng_oup; event.klasse = ENG_EVENT_UP_RIGHT; ENG_SendEvent(&event);
						} else {
							pyu = -1; pxu = 1;
						}
						break;
					case ENG_ACTION_RIGHT:
						if (eng_op_env.pointer_mode == ENG_POINTER_MODE_OFF || eng_op_env.pointer_mode == ENG_POINTER_MODE_PHYSICAL) {
							event.object = eng_oup; event.klasse = ENG_EVENT_RIGHT; ENG_SendEvent(&event);
						} else if (eng_op_env.pointer_mode == ENG_POINTER_MODE_SIMULATED) {
							pxu = 1;
						}
						break;
					case ENG_ACTION_DOWN_RIGHT:
						if (eng_op_env.pointer_mode == ENG_POINTER_MODE_OFF || eng_op_env.pointer_mode == ENG_POINTER_MODE_PHYSICAL) {
							event.object = eng_oup; event.klasse = ENG_EVENT_DOWN_RIGHT; ENG_SendEvent(&event);
						} else {
							pyu = 1; pxu = 1;
						}
						break;
					case ENG_ACTION_DOWN:
						if (eng_op_env.pointer_mode == ENG_POINTER_MODE_OFF || eng_op_env.pointer_mode == ENG_POINTER_MODE_PHYSICAL) {
							event.object = eng_oup; event.klasse = ENG_EVENT_DOWN; ENG_SendEvent(&event);
						} else {
							pyu = 1;
						}
						break;
					case ENG_ACTION_DOWN_LEFT:
						if (eng_op_env.pointer_mode == ENG_POINTER_MODE_OFF || eng_op_env.pointer_mode == ENG_POINTER_MODE_PHYSICAL) {
							event.object = eng_oup; event.klasse = ENG_EVENT_DOWN_LEFT; ENG_SendEvent(&event);
						} else {
							pyu = 1; pxu = -1;
						}
						break;
					case ENG_ACTION_LEFT:
						if (eng_op_env.pointer_mode == ENG_POINTER_MODE_OFF || eng_op_env.pointer_mode == ENG_POINTER_MODE_PHYSICAL) {
							event.object = eng_oup; event.klasse = ENG_EVENT_LEFT; ENG_SendEvent(&event);
						} else {
							pxu = -1;
						}
						break;
					case ENG_ACTION_UP_LEFT:
						if (eng_op_env.pointer_mode == ENG_POINTER_MODE_OFF || eng_op_env.pointer_mode == ENG_POINTER_MODE_PHYSICAL) {
							event.object = eng_oup; event.klasse = ENG_EVENT_UP_LEFT; ENG_SendEvent(&event);
						} else {
							pyu = -1; pxu = -1;
						}
						break;
					case ENG_ACTION_SELECT:
						if (eng_op_env.pointer_mode != ENG_POINTER_MODE_OFF) {
							if (eng_pressed_object == NULL) {
								eng_pressed_object = eng_oup;
								/* Only generate events for enabled objects */
								if (ENG_GetEnability(eng_oup)) {
									event.object = eng_oup; event.klasse = ENG_EVENT_PRESS; ENG_SendEvent(&event);
									event.object = eng_oup; event.klasse = ENG_EVENT_SELECT; ENG_SendEvent(&event);
									/* For draggable [enabled] objects, calculate the drag points */
									if (eng_oup->draggable) {
										/* Get the current pointer coordinates */
										ENG_GetCanvasOffset(eng_lyr_pointer, &mx, &my);
										if (eng_oup->drag_point == ENG_DRAG_POINT_ANYWHERE) {
											press_point_x = mx - px - eng_oup->x;
											press_point_y = my - py - eng_oup->y;
										} else if (eng_oup->drag_point == ENG_DRAG_POINT_MIDDLE) {
											press_point_x = eng_oup->w / 2;
											press_point_y = eng_oup->h / 2;
										}
									}
								}
							} else {	/* Something is already being pressed */
								/* Only generate events for enabled objects */
								if (eng_pressed_object == eng_oup && ENG_GetEnability(eng_oup)) {
									event.object = eng_oup; event.klasse = ENG_EVENT_SELECT; ENG_SendEvent(&event);
								}
							}
						} else {
							event.object = eng_oup; event.klasse = ENG_EVENT_SELECT; ENG_SendEvent(&event);
						}
						break;
					case ENG_ACTION_SELECT | 0x80:
						if (eng_op_env.pointer_mode != ENG_POINTER_MODE_OFF) {
							if (eng_pressed_object == eng_oup) {
								/* Only generate events for enabled objects */
								if (ENG_GetEnability(eng_oup)) {
									event.object = eng_oup; event.klasse = ENG_EVENT_CLICK; ENG_SendEvent(&event);
								}
							} else {
								/* Only generate events for enabled objects */
								if (ENG_GetEnability(eng_pressed_object)) {
									event.object = eng_oup; event.klasse = ENG_EVENT_RELEASE; ENG_SendEvent(&event);
								}
							}
							eng_pressed_object = NULL;
						}
						break;
					case ENG_ACTION_SHUTDOWN:
						event.object = eng_oup; event.klasse = ENG_EVENT_SHUTDOWN; ENG_SendEvent(&event);
						break;
					case ENG_ACTION_TOGGLE_FULLSCREEN:
						event.object = eng_oup; event.klasse = ENG_EVENT_TOGGLE_FULLSCREEN; ENG_SendEvent(&event);
						break;
					case ENG_ACTION_MODIFIER1:
						event.object = eng_oup; event.klasse = ENG_EVENT_MODIFIER1; ENG_SendEvent(&event);
						break;
					case ENG_ACTION_MODIFIER2:
						event.object = eng_oup; event.klasse = ENG_EVENT_MODIFIER2; ENG_SendEvent(&event);
						break;
				}
			}
		}
	} while (actionid != ENG_UNDEFINED);

	/* Manage pointer input if required */
	if (eng_op_env.pointer_mode == ENG_POINTER_MODE_SIMULATED || eng_op_env.pointer_mode == ENG_POINTER_MODE_NAVIGATION) {
		ENG_ManagePointerInput(pxu, pyu);
	}
				
}

/***************************************************************************
 * Manage Pointer Input                                                    *
 ***************************************************************************/
/* This manages pointer input for certain pointer modes */

void ENG_ManagePointerInput(int pxu, int pyu) {
	static int lastpxu = 0, lastpyu = 0, unitcount = 0;
	int xmov, ymov;
	
	if (eng_op_env.pointer_mode == ENG_POINTER_MODE_SIMULATED) {
		/* Is the x or y unit the same as last time? */
		if ((pxu != 0 && pxu == lastpxu) || (pyu != 0 && pyu == lastpyu)) {
			unitcount++;
			/* If the count reaches a threshold then change gear */
			if (unitcount < eng_op_env.pointer_move_unit_threshold) {
				xmov = pxu * eng_op_env.pointer_move_unit_low;
				ymov = pyu * eng_op_env.pointer_move_unit_low;
			} else {
				xmov = pxu * eng_op_env.pointer_move_unit_high;
				ymov = pyu * eng_op_env.pointer_move_unit_high;
			}
		} else {
			unitcount = 0;
			xmov = pxu * eng_op_env.pointer_move_unit_low;
			ymov = pyu * eng_op_env.pointer_move_unit_low;
		}
		lastpxu = pxu; lastpyu = pyu;
			
		/* Now move the pointer */
		eng_lyr_pointer->x += xmov;
		eng_lyr_pointer->y += ymov;

		/* Keep it within the bounds of the screen */
		if (eng_lyr_pointer->x < 0) {
			eng_lyr_pointer->x = 0;
		} else if (eng_lyr_pointer->x > eng_op_env.screen_w - 1) {
			eng_lyr_pointer->x = eng_op_env.screen_w - 1;
		}
		if (eng_lyr_pointer->y < 0) {
			eng_lyr_pointer->y = 0;
		} else if (eng_lyr_pointer->y > eng_op_env.screen_h - 1) {
			eng_lyr_pointer->y = eng_op_env.screen_h - 1;
		}

	} else if (eng_op_env.pointer_mode == ENG_POINTER_MODE_NAVIGATION) {
		/* Not implemented yet temp temp */
	}

}

/***************************************************************************
 * Send Event                                                              *
 ***************************************************************************/
/* This first sends events to the ENG event processor to service system
   object events and then passes them on the the user's event processor */

/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_SendEvent(ENG_Event *event) {
	int retval = ENG_ERROR_NONE;

	if (eng_started == FALSE) {
		sprintf(eng_last_error, "%s: ENG has not been started yet", __func__);
		retval = ENG_ERROR_ENG_NOT_YET_STARTED;
	} else if (event == NULL || sizeof(*event) != sizeof(ENG_Event)) {
		sprintf(eng_last_error, "%s: event is invalid", __func__);
		retval = ENG_ERROR_INVALID_EVENT;
	} else {
		#ifdef DEBUG_EVENTS
			ENG_ShowEvent(event);
		#endif
		ENG_EventProcessor(event);			/* System object events */
		eng_op_env.event_processor(event);	/* All other events */
	}
	
	return retval;
}

/***************************************************************************
 * Event Processor                                                         *
 ***************************************************************************/
/* This processes events for system objects */

void ENG_EventProcessor(ENG_Event *event) {



}

/***************************************************************************
 * Show Event                                                              *
 ***************************************************************************/

int ENG_ShowEvent(ENG_Event *event) {
	int retval = ENG_ERROR_NONE;

	if (event == NULL || sizeof(*event) != sizeof(ENG_Event)) {
		sprintf(eng_last_error, "%s: event is invalid", __func__);
		retval = ENG_ERROR_INVALID_EVENT;
	} else {
		if (event->klasse != ENG_EVENT_PRE_RENDER &&	/* Filter out these repeating events */
			event->klasse != ENG_EVENT_POST_RENDER &&
			event->klasse != ENG_EVENT_CYCLE) {

			switch (event->klasse) {
				case ENG_EVENT_OVER:
					printf("%s: ENG_EVENT_OVER", __func__);
					break;
				case ENG_EVENT_OUT:
					printf("%s: ENG_EVENT_OUT", __func__);
					break;
				case ENG_EVENT_PRESS:
					printf("%s: ENG_EVENT_PRESS", __func__);
					break;
				case ENG_EVENT_CLICK:
					printf("%s: ENG_EVENT_CLICK", __func__);
					break;
				case ENG_EVENT_RELEASE:
					printf("%s: ENG_EVENT_RELEASE", __func__);
					break;
				case ENG_EVENT_DRAG:
					printf("%s: ENG_EVENT_DRAG", __func__);
					break;
				case ENG_EVENT_QUITTING:
					printf("%s: ENG_EVENT_QUITTING", __func__);
					break;
				case ENG_EVENT_POST_LOAD_OBJECTS:
					printf("%s: ENG_EVENT_POST_LOAD_OBJECTS", __func__);
					break;
				case ENG_EVENT_PRE_RENDER:
					printf("%s: ENG_EVENT_PRE_RENDER", __func__);
					break;
				case ENG_EVENT_POST_RENDER:
					printf("%s: ENG_EVENT_POST_RENDER", __func__);
					break;
				case ENG_EVENT_CYCLE:
					printf("%s: ENG_EVENT_CYCLE", __func__);
					break;
				case ENG_EVENT_POST_EFFECT:
					printf("%s: ENG_EVENT_POST_EFFECT", __func__);
					break;
				case ENG_EVENT_UP:
					printf("%s: ENG_EVENT_UP", __func__);
					break;
				case ENG_EVENT_UP_RIGHT:
					printf("%s: ENG_EVENT_UP_RIGHT", __func__);
					break;
				case ENG_EVENT_RIGHT:
					printf("%s: ENG_EVENT_RIGHT", __func__);
					break;
				case ENG_EVENT_DOWN_RIGHT:
					printf("%s: ENG_EVENT_DOWN_RIGHT", __func__);
					break;
				case ENG_EVENT_DOWN:
					printf("%s: ENG_EVENT_DOWN", __func__);
					break;
				case ENG_EVENT_DOWN_LEFT:
					printf("%s: ENG_EVENT_DOWN_LEFT", __func__);
					break;
				case ENG_EVENT_LEFT:
					printf("%s: ENG_EVENT_LEFT", __func__);
					break;
				case ENG_EVENT_UP_LEFT:
					printf("%s: ENG_EVENT_UP_LEFT", __func__);
					break;
				case ENG_EVENT_SELECT:
					printf("%s: ENG_EVENT_SELECT", __func__);
					break;
				case ENG_EVENT_SHUTDOWN:
					printf("%s: ENG_EVENT_SHUTDOWN", __func__);
					break;
				case ENG_EVENT_TOGGLE_FULLSCREEN:
					printf("%s: ENG_EVENT_TOGGLE_FULLSCREEN", __func__);
					break;
				case ENG_EVENT_MODIFIER1:
					printf("%s: ENG_EVENT_MODIFIER1", __func__);
					break;
				case ENG_EVENT_MODIFIER2:
					printf("%s: ENG_EVENT_MODIFIER2", __func__);
					break;
				default:
					printf("%s: Unknown", __func__);
					break;
			}
			printf(" (%i), id=%i object=%p\n", event->klasse, event->object->id, event->object);
		}
	}

	return retval;	
}




