/* ENG is Not a Game (c) Thunor 2008 <thunorsif@hotmail.com>
www.thunor.org.uk

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */


/* Includes */
#include "ENG_engine.h"


/* Defines */
/*
#define DEBUG_OP_ENV
*/

#define ENG_MAX_DELAY 100


/***************************************************************************
 * Initialise Operating Environment                                        *
 ***************************************************************************/

void ENG_InitOpEnv(void) {
}

/***************************************************************************
 * Set Operating Environment - Image Directory                             *
 ***************************************************************************/
/* On entry: image_dir points to the dir where all images are to be found */
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_SetOpEnvImageDir(char *image_dir) {
	int retval = ENG_ERROR_NONE;
	
	#ifdef DEBUG_OP_ENV
		printf("%s: Setting\n", __func__);
	#endif
	/* No validation is done */
	/* Store new value(s) */
	strcpy(eng_op_env.image_dir, image_dir);
	
	return retval;
}

/***************************************************************************
 * Set Operating Environment - Objects File                                *
 ***************************************************************************/
/* On entry: objects_file points to the file that holds all the objects */
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_SetOpEnvObjectsFile(char *objects_file) {
	int retval = ENG_ERROR_NONE;
	ENG_OpEnv local_op_env;
	
	#ifdef DEBUG_OP_ENV
		printf("%s: Setting\n", __func__);
	#endif
	/* Validate new value(s) */
	strcpy(local_op_env.objects_file, objects_file);
	if (!(retval = ENG_ValidateOpEnv(&local_op_env, ENG_VALIDATE_OBJECTS_FILE))) {
		/* Store new value(s) */
		strcpy(eng_op_env.objects_file, objects_file);
	}

	return retval;
}

/***************************************************************************
 * Set Operating Environment - Config File                                 *
 ***************************************************************************/
/* On entry: config_file points to the applications's configuration file */
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_SetOpEnvConfigFile(char *config_file) {
	int retval = ENG_ERROR_NONE;
	
	#ifdef DEBUG_OP_ENV
		printf("%s: Setting\n", __func__);
	#endif
	/* Save the config file (ignoring errors) */
	if (strcmp(eng_op_env.config_file, "") != 0) {
		ENG_SaveConfig();
		/* Free all the existing config options */
		ENG_FreeAllConfigOptions();
	}
	/* No validation is done */
	/* Store new value(s) */
	strcpy(eng_op_env.config_file, config_file);
	/* Load the new config file */
	if (strcmp(eng_op_env.config_file, "") != 0) retval = ENG_LoadConfig();
	
	return retval;
}

/***************************************************************************
 * Set Operating Environment - Screen                                      *
 ***************************************************************************/
/* If SDL is being restarted then the screen will not be resized now; only
   the dimensions will be saved in eng_op_env otherwise the current display
   surface (the screen) is resized and if it fails an attempt is made to
   restore the original. If that fails then the engine will shutdown.
   In any case, an error code will be returned */

/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_SetOpEnvScreen(int screen_w, int screen_h) {
	int retval = ENG_ERROR_NONE;
	Uint8 bpp;
	Uint32 flags;
	
	#ifdef DEBUG_OP_ENV
		printf("%s: Setting\n", __func__);
	#endif
	/* No validation is done */
	/* If a screen surface exists then resize screen now */
	if (eng_screen != NULL) {
		/* First backup current screen settings */
		bpp = eng_screen->format->BitsPerPixel;
		flags = eng_screen->flags;
		/* Reset video mode now */
		printf("%s: Setting video mode %ix%i %ibpp\n", __func__, screen_w, screen_h, bpp);
		eng_screen = SDL_SetVideoMode(screen_w, screen_h, bpp, flags);
		if (eng_screen == NULL) {
			sprintf(eng_last_error, "%s: Cannot set video mode: %s", __func__, SDL_GetError());
			retval = ENG_ERROR_CANNOT_SET_VIDEO_MODE;
			/* Restore original video mode */
			printf("%s: Restoring original video mode %ix%i %ibpp\n", __func__, eng_op_env.screen_w, eng_op_env.screen_h, bpp);
			eng_screen = SDL_SetVideoMode(eng_op_env.screen_w, eng_op_env.screen_h, bpp, flags);
			if (eng_screen == NULL) {
				sprintf(eng_last_error, "%s: Cannot set video mode: %s", __func__, SDL_GetError());
				retval = ENG_ERROR_CANNOT_SET_VIDEO_MODE;
				printf("%s: Shutting engine down: No display surface\n", __func__);
				ENG_Quit();	/* No point in continuing with no screen :) */
			}
		} else {
			/* Store new value(s) */
			eng_op_env.screen_w = screen_w;
			eng_op_env.screen_h = screen_h;
			/* Free blend screen surface */
			if (eng_blend_screen != NULL) SDL_FreeSurface(eng_blend_screen);
			/* Duplicate the screen surface for making alpha rectangles only */
			eng_blend_screen = SDL_CreateRGBSurface(eng_screen->flags | SDL_SRCALPHA, eng_op_env.screen_w, eng_op_env.screen_h, eng_screen->format->BitsPerPixel, eng_screen->format->Rmask, eng_screen->format->Gmask, eng_screen->format->Bmask, eng_screen->format->Amask);
			if (eng_blend_screen == NULL) {
				sprintf(eng_last_error, "%s: Cannot create blend surface: %s", __func__, SDL_GetError());
				retval = ENG_ERROR_CANNOT_CREATE_SURFACE;
				printf("%s: Shutting engine down: No blend surface\n", __func__);
				ENG_Quit();	/* No point in continuing with no blend screen */
			}
		}
	} else {
		/* Store new value(s) */
		eng_op_env.screen_w = screen_w;
		eng_op_env.screen_h = screen_h;
	}
	
	return retval;
}

/***************************************************************************
 * Set Operating Environment - Canvas                                      *
 ***************************************************************************/
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_SetOpEnvCanvas(int canvas_w, int canvas_h) {
	int retval = ENG_ERROR_NONE;
	
	#ifdef DEBUG_OP_ENV
		printf("%s: Setting\n", __func__);
	#endif
	/* No validation is done */
	/* Store new value(s) */
	eng_op_env.canvas_w = canvas_w;
	eng_op_env.canvas_h = canvas_h;
	/* DO SOMETHING HERE MAYBE temp temp */
	/* DO SOMETHING HERE MAYBE temp temp */
	/* DO SOMETHING HERE MAYBE temp temp */
	
	return retval;
}

/***************************************************************************
 * Set Operating Environment - Viewport                                    *
 ***************************************************************************/
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_SetOpEnvViewport(int viewport_x, int viewport_y, int viewport_w, int viewport_h, int viewport_screen_x, int viewport_screen_y) {
	int retval = ENG_ERROR_NONE;
	
	#ifdef DEBUG_OP_ENV
		printf("%s: Setting\n", __func__);
	#endif
	/* No validation is done */
	/* Store new value(s) */
	eng_op_env.viewport_x = viewport_x;
	eng_op_env.viewport_y = viewport_y;
	eng_op_env.viewport_w = viewport_w;
	eng_op_env.viewport_h = viewport_h;
	eng_op_env.viewport_screen_x = viewport_screen_x;
	eng_op_env.viewport_screen_y = viewport_screen_y;
	/* Update ENG_LYR_SCREEN */
	eng_lyr_screen->x = eng_op_env.viewport_x - eng_op_env.viewport_screen_x;
	eng_lyr_screen->y = eng_op_env.viewport_y - eng_op_env.viewport_screen_y;
	/* Update ENG_LYR_VIEWPORT */
	eng_lyr_viewport->x = eng_op_env.viewport_screen_x;
	eng_lyr_viewport->y = eng_op_env.viewport_screen_y;
	eng_lyr_viewport->w = eng_op_env.viewport_w;
	eng_lyr_viewport->h = eng_op_env.viewport_h;
	
	return retval;
}

/***************************************************************************
 * Set Operating Environment - Title                                       *
 ***************************************************************************/
/* On entry: title points to a title for window managers */
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_SetOpEnvTitle(char *title) {
	int retval = ENG_ERROR_NONE;
	
	#ifdef DEBUG_OP_ENV
		printf("%s: Setting\n", __func__);
	#endif
	/* No validation is done */
	/* Store new value(s) */
	strcpy(eng_op_env.title, title);
	/* Set window manager title */
	if (eng_screen != NULL) SDL_WM_SetCaption(eng_op_env.title, "");
	
	return retval;
}

/***************************************************************************
 * Set Operating Environment - Full Screen                                 *
 ***************************************************************************/
/* On entry: fullscreen is true or false */
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_SetOpEnvFullScreen(int fullscreen) {
	int retval = ENG_ERROR_NONE;
	ENG_OpEnv local_op_env;
	
	#ifdef DEBUG_OP_ENV
		printf("%s: Setting\n", __func__);
	#endif
	/* Validate new value(s) */
	local_op_env.fullscreen = fullscreen;
	if (!(retval = ENG_ValidateOpEnv(&local_op_env, ENG_VALIDATE_FULLSCREEN))) {
		if ((eng_screen != NULL) && (SDL_WM_ToggleFullScreen(eng_screen)) == 0) {
			sprintf(eng_last_error, "%s: Cannot toggle fullscreen: %s", __func__, SDL_GetError());
			retval = ENG_ERROR_CANNOT_TOGGLE_FULLSCREEN;
		} else {
			/* Store new value(s) */
			eng_op_env.fullscreen = fullscreen;
		}
	}
	
	return retval;
}

/***************************************************************************
 * Set Operating Environment - System Pointer                              *
 ***************************************************************************/
/* On entry: systempointer is true or false */
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_SetOpEnvSystemPointer(int systempointer) {
	int retval = ENG_ERROR_NONE;
	ENG_OpEnv local_op_env;
	
	#ifdef DEBUG_OP_ENV
		printf("%s: Setting\n", __func__);
	#endif
	/* Validate new value(s) */
	local_op_env.systempointer = systempointer;
	if (!(retval = ENG_ValidateOpEnv(&local_op_env, ENG_VALIDATE_SYSTEMPOINTER))) {
		/* Store new value(s) */
		eng_op_env.systempointer = systempointer;
		if (eng_screen != NULL) SDL_ShowCursor(systempointer);
	}
	
	return retval;
}

/***************************************************************************
 * Set Operating Environment - Pointer                                     *
 ***************************************************************************/
/* On entry: pointer's visibility is true or false
			 pointer_mode is one of the modes as listed in ENG.h */
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_SetOpEnvPointer(int pointer, int pointer_mode) {
	int retval = ENG_ERROR_NONE;
	ENG_OpEnv local_op_env;

	#ifdef DEBUG_OP_ENV
		printf("%s: Setting\n", __func__);
	#endif
	/* Validate new value(s) */
	local_op_env.pointer = pointer;
	local_op_env.pointer_mode = pointer_mode;
	if (!(retval = ENG_ValidateOpEnv(&local_op_env, ENG_VALIDATE_POINTER))) {
		if (!(retval = ENG_ValidateOpEnv(&local_op_env, ENG_VALIDATE_POINTER_MODE))) {
			/* Store new value(s) */
			eng_op_env.pointer = pointer;
			eng_op_env.pointer_mode = pointer_mode;
			/* DO SOMETHING HERE MAYBE temp temp */
			/* DO SOMETHING HERE MAYBE temp temp */
			if (eng_object_count > 0) eng_lyr_pointer->visible = pointer;
		}
	}
	
	return retval;
}

/***************************************************************************
 * Set Operating Environment - Pointer Move Units                          *
 ***************************************************************************/
/* On entry: pointer_move_unit_low is the unit the pointer should move
			 unaccelerated
			 pointer_move_unit_high is the unit the pointer should move
			 accelerated
			 pointer_move_unit_threshold is the number of units marking
			 the low/high threshold */
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_SetOpEnvPointerMoveUnits(int pointer_move_unit_low, int pointer_move_unit_high, int pointer_move_unit_threshold) {
	int retval = ENG_ERROR_NONE;

	#ifdef DEBUG_OP_ENV
		printf("%s: Setting\n", __func__);
	#endif
	/* No validation is done */
	/* Store new value(s) */
	eng_op_env.pointer_move_unit_low = pointer_move_unit_low;
	eng_op_env.pointer_move_unit_high = pointer_move_unit_high;
	eng_op_env.pointer_move_unit_threshold = pointer_move_unit_threshold;
	/* DO SOMETHING HERE MAYBE temp temp */
	/* DO SOMETHING HERE MAYBE temp temp */
	/* DO SOMETHING HERE MAYBE temp temp */
	
	return retval;
}

/***************************************************************************
 * Set Operating Environment - Delay And Cycle Limit                       *
 ***************************************************************************/
/* On entry: delay in ms will be passed to SDL_Delay in ENG_Woden if there
			 is no cycle_limit
			 cycle_limit is used along with a timer to set a flag which
			 enables ENG_Woden to proceed */
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_SetOpEnvDelayAndCycleLimit(int delay, int cycle_limit) {
	int retval = ENG_ERROR_NONE;
	ENG_OpEnv local_op_env;

	#ifdef DEBUG_OP_ENV
		printf("%s: Setting\n", __func__);
	#endif
	/* Validate new value(s) */
	local_op_env.delay = delay;
	local_op_env.cycle_limit = cycle_limit;
	if (!(retval = ENG_ValidateOpEnv(&local_op_env, ENG_VALIDATE_DELAY))) {
		if (!(retval = ENG_ValidateOpEnv(&local_op_env, ENG_VALIDATE_CYCLE_LIMIT))) {
			/* Store new value(s) */
			eng_op_env.delay = delay;
			eng_op_env.cycle_limit = cycle_limit;
		}
	}
	
	return retval;
}

/***************************************************************************
 * Set Operating Environment - Event Processor                             *
 ***************************************************************************/
/* On entry: event_processor is a pointer to a function that receives events */
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_SetOpEnvEventProcessor(void (*event_processor)(ENG_Event *event)) {
	int retval = ENG_ERROR_NONE;
	ENG_OpEnv local_op_env;
	
	#ifdef DEBUG_OP_ENV
		printf("%s: Setting\n", __func__);
	#endif
	/* Validate new value(s) */
	local_op_env.event_processor = event_processor;
	if (!(retval = ENG_ValidateOpEnv(&local_op_env, ENG_VALIDATE_EVENT_PROCESSOR))) {
		/* Store new value(s) */
		eng_op_env.event_processor = event_processor;
	}
	
	return retval;
}

/***************************************************************************
 * Validate Operating Environment                                          *
 ***************************************************************************/
/* On entry: op_env points to the operating environment data
			 which declares what will be validated */
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_ValidateOpEnv(ENG_OpEnv *op_env, int which) {
	int retval = ENG_ERROR_NONE;
	
	#ifdef DEBUG_OP_ENV
		printf("%s: Validating\n", __func__);
	#endif
	/* I've decided not to validate the coordinates and dimensions
	   because of the many different combinations there could be.
	   In any case if they're unset they'll all be 0 */
	if (which == ENG_VALIDATE_OBJECTS_FILE || which == ENG_VALIDATE_ALL) {
		if (op_env->objects_file[0] == 0) {
			sprintf(eng_last_error, "%s: objects_file is empty", __func__);
			retval = ENG_ERROR_INVALID_OP_ENV;
		}
	}
	/*if (which == ENG_VALIDATE_CONFIG_FILE || which == ENG_VALIDATE_ALL) { redundant
		if (op_env->config_file[0] == 0) {
			sprintf(eng_last_error, "%s: config_file is empty", __func__);
			retval = ENG_ERROR_INVALID_OP_ENV;
		}
	}*/
	if (which == ENG_VALIDATE_FULLSCREEN || which == ENG_VALIDATE_ALL) {
		if (op_env->fullscreen != 0 && op_env->fullscreen != 1) {
			sprintf(eng_last_error, "%s: fullscreen is invalid", __func__);
			retval = ENG_ERROR_INVALID_OP_ENV;
		}
	}
	if (which == ENG_VALIDATE_SYSTEMPOINTER || which == ENG_VALIDATE_ALL) {
		if (op_env->systempointer != 0 && op_env->systempointer != 1) {
			sprintf(eng_last_error, "%s: systempointer is invalid", __func__);
			retval = ENG_ERROR_INVALID_OP_ENV;
		}
	}
	if (which == ENG_VALIDATE_POINTER || which == ENG_VALIDATE_ALL) {
		if (op_env->pointer != 0 && op_env->pointer != 1) {
			sprintf(eng_last_error, "%s: pointer is invalid", __func__);
			retval = ENG_ERROR_INVALID_OP_ENV;
		}
	}
	if (which == ENG_VALIDATE_POINTER_MODE || which == ENG_VALIDATE_ALL) {
		if (op_env->pointer_mode < ENG_POINTER_MODE_OFF || op_env->pointer_mode > ENG_POINTER_MODE_NAVIGATION) {
			sprintf(eng_last_error, "%s: pointer_mode is invalid", __func__);
			retval = ENG_ERROR_INVALID_OP_ENV;
		}
	}
	if (which == ENG_VALIDATE_DELAY || which == ENG_VALIDATE_ALL) {
		if (op_env->delay < 0 || op_env->delay > ENG_MAX_DELAY) {
			sprintf(eng_last_error, "%s: delay is invalid", __func__);
			retval = ENG_ERROR_INVALID_OP_ENV;
		}
	}
	if (which == ENG_VALIDATE_CYCLE_LIMIT || which == ENG_VALIDATE_ALL) {
		if (op_env->cycle_limit != 0 && op_env->cycle_limit != 16 && op_env->cycle_limit != 20 && op_env->cycle_limit != 25 && op_env->cycle_limit != 33 && op_env->cycle_limit != 50 && op_env->cycle_limit != 100) {
			sprintf(eng_last_error, "%s: cycle_limit is invalid", __func__);
			retval = ENG_ERROR_INVALID_OP_ENV;
		}
	}
	if (which == ENG_VALIDATE_EVENT_PROCESSOR || which == ENG_VALIDATE_ALL) {
		if (op_env->event_processor == NULL) {
			sprintf(eng_last_error, "%s: event_processor() is unset", __func__);
			retval = ENG_ERROR_INVALID_OP_ENV;
		}
	}
	#ifdef DEBUG_OP_ENV
		printf("op_env :-\n");
		printf("image_dir=%s\n", op_env->image_dir);
		printf("objects_file=%s\n", op_env->objects_file);
		printf("config_file=%s\n", op_env->config_file);
		printf("screen_w=%i\n", op_env->screen_w);
		printf("screen_h=%i\n", op_env->screen_h);
		printf("canvas_w=%i\n", op_env->canvas_w);
		printf("canvas_h=%i\n", op_env->canvas_h);
		printf("viewport_x=%i\n", op_env->viewport_x);
		printf("viewport_y=%i\n", op_env->viewport_y);
		printf("viewport_w=%i\n", op_env->viewport_w);
		printf("viewport_h=%i\n", op_env->viewport_h);
		printf("viewport_screen_x=%i\n", op_env->viewport_screen_x);
		printf("viewport_screen_y=%i\n", op_env->viewport_screen_y);
		printf("title=%s\n", op_env->title);
		printf("fullscreen=%i\n", op_env->fullscreen);
		printf("systempointer=%i\n", op_env->systempointer);
		printf("pointer=%i\n", op_env->pointer);
		printf("pointer_mode=%i\n", op_env->pointer_mode);
		printf("pointer_move_unit_low=%i\n", op_env->pointer_move_unit_low);
		printf("pointer_move_unit_high=%i\n", op_env->pointer_move_unit_high);
		printf("pointer_move_unit_threshold=%i\n", op_env->pointer_move_unit_threshold);
		printf("delay=%i\n", op_env->delay);
		printf("cycle_limit=%i\n", op_env->cycle_limit);
		printf("event_processor()=%p\n", op_env->event_processor);
	#endif

	return retval;
}



