/* ENG is Not a Game (c) Thunor 2008 <thunorsif@hotmail.com>
www.thunor.org.uk

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */


/* Includes */
#include "ENG_engine.h"


/* Defines */
#define ENG_IMG_ICON "icon.bmp"


/* Variables */
SDL_Surface *eng_icon = NULL;
int eng_joystick_count = 0;


/* Function prototypes */
Uint32 ENG_Timer(Uint32 interval, void *param);
char *ENG_GetJoystickName(int joyid);


/***************************************************************************
 * Initialise SDL                                                          *
 ***************************************************************************/

void ENG_InitSDL(void) {
	eng_screen = NULL;
	eng_blend_screen = NULL;
}

/***************************************************************************
 * Start SDL                                                               *
 ***************************************************************************/
/*  On exit: returns ENG_ERROR_NONE on success else
			 any one of the error codes in ENG.h */

int ENG_StartSDL(void) {
	int bpp = 0, flags = 0, count;
	SDL_Event event;
	#ifdef PLATFORM_PC
		char temp[ENG_PATH_LENGTH + ENG_FILE_LENGTH];
		SDL_Surface *icon;
		Uint32 rmask, gmask, bmask, amask;
	#endif
	
	/* Initialise flags. Software surfaces are used because of the engine's
	   capability of using alpha and I haven't yet found a video device that
	   offers hardware accelerated alpha blending */
	#if defined(PLATFORM_PC)
		bpp = 32;
		flags |= SDL_SWSURFACE;
		flags |= SDL_ANYFORMAT;
		flags |= eng_op_env.fullscreen * SDL_FULLSCREEN;
	#elif defined(PLATFORM_ZAURUS)
		bpp = 16;
		flags |= SDL_SWSURFACE;
		flags |= SDL_FULLSCREEN;
	#elif defined(PLATFORM_GP2X)
		bpp = 16;
		flags |= SDL_SWSURFACE;
		flags |= SDL_FULLSCREEN;
	#endif

	/* Initialise SDL */
	if (SDL_Init(SDL_INIT_EVERYTHING)) {
		sprintf(eng_last_error, "%s: Cannot initialise SDL: %s", __func__, SDL_GetError());
		return ENG_ERROR_CANNOT_INIT_SDL;
	}

	#ifdef PLATFORM_PC
		/* Set display window icon
		   Note: All errors here are not serious and are sent straight to the terminal */
		strcpy(temp, eng_op_env.image_dir);
		if (temp[strlen(temp) - 1] != '/') strcat(temp, "/");
		strcat(temp, ENG_IMG_ICON);
		if ((icon = SDL_LoadBMP(temp)) == NULL) {
			printf("%s: %s not found: %s\n", __func__, temp, SDL_GetError());
		} else {
			/* Set the colour key for the new surface */
			if ((SDL_SetColorKey(icon, SDL_SRCCOLORKEY, SDL_MapRGB(icon->format, ENG_COLOUR_KEY_RED, ENG_COLOUR_KEY_GREEN, ENG_COLOUR_KEY_BLUE))) < 0) {
				printf("%s: Cannot set display window icon surface colour key: %s\n", __func__, SDL_GetError());
			}
			#if SDL_BYTEORDER == SDL_BIG_ENDIAN
				rmask = 0xff000000; gmask = 0x00ff0000; bmask = 0x0000ff00; amask = 0x000000ff;
			#else
				rmask = 0x000000ff; gmask = 0x0000ff00; bmask = 0x00ff0000; amask = 0xff000000;
			#endif
			if ((eng_icon = SDL_CreateRGBSurface(SDL_SWSURFACE, 32, 32, 32, rmask, gmask, bmask, amask)) == NULL) {
				printf("%s: Cannot create display window icon surface: %s\n", __func__, SDL_GetError());
			} else {
				if ((SDL_BlitSurface(icon, NULL, eng_icon, NULL)) < 0) {
					printf("%s: Cannot blit display window icon surface: %s\n", __func__, SDL_GetError());
				} else {
					SDL_WM_SetIcon(eng_icon, NULL);
				}
			}
			SDL_FreeSurface(icon);
		}
	#endif
	
	/* Set video mode */
	printf("%s: Setting video mode %ix%i %ibpp\n", __func__, eng_op_env.screen_w, eng_op_env.screen_h, bpp);
	eng_screen = SDL_SetVideoMode(eng_op_env.screen_w, eng_op_env.screen_h, bpp, flags);
	if (eng_screen == NULL) {
		sprintf(eng_last_error, "%s: Cannot set video mode: %s", __func__, SDL_GetError());
		return ENG_ERROR_CANNOT_SET_VIDEO_MODE;
	}

	/* Duplicate the screen surface for making alpha rectangles only */
	eng_blend_screen = SDL_CreateRGBSurface(flags | SDL_SRCALPHA, eng_op_env.screen_w, eng_op_env.screen_h, bpp, eng_screen->format->Rmask, eng_screen->format->Gmask, eng_screen->format->Bmask, eng_screen->format->Amask);
	if (eng_blend_screen == NULL) {
		sprintf(eng_last_error, "%s: Cannot create blend surface: %s", __func__, SDL_GetError());
		return ENG_ERROR_CANNOT_CREATE_SURFACE;
	}
	
	#ifdef PLATFORM_PC
		/* Set display window title here */
		SDL_WM_SetCaption(eng_op_env.title, eng_op_env.title);
	#endif

	/* Set physical mouse pointer state here */
	SDL_ShowCursor(eng_op_env.systempointer);

	/* Attempt to open the number of joysticks reported (or ENG_MAX_JOYSTICKS)
	   and any that don't open will have NULL pointers within eng_joysticks */
	eng_joystick_count = SDL_NumJoysticks();
	if (eng_joystick_count > ENG_MAX_JOYSTICKS) eng_joystick_count = ENG_MAX_JOYSTICKS;
	if (eng_joystick_count > 0) {
		SDL_JoystickEventState(SDL_ENABLE);	/* Enable joystick event processing */	
		for (count = 0; count < eng_joystick_count; count++) {
			/* Nullify joystick pointer so we can guarantee its validity */
			eng_joysticks[count] = NULL;
			/* Open joystick */
			eng_joysticks[count] = SDL_JoystickOpen(count);
			if (eng_joysticks[count] != NULL) {
				printf("%s: Joystick opened: %i:%s\n", __func__, count, ENG_GetJoystickName(count));
			} else {
				printf("%s: Couldn't open joystick: %i:%s\n", __func__, count, ENG_GetJoystickName(count));
			}
		}
		/* Flush events as opening a joystick releases all the buttons */
		while (SDL_PollEvent(&event));
	} else {
		#if defined(PLATFORM_PC)
			printf("%s: No joysticks found\n", __func__);
		#elif defined(PLATFORM_GP2X)
			sprintf(eng_last_error, "%s: Cannot open any joystick\n", __func__);
			return ENG_ERROR_CANNOT_OPEN_JOYSTICK;
		#endif
	}

	/* Initiate a 10ms timer */
	eng_timer_id = SDL_AddTimer(10, ENG_Timer, NULL);

	return ENG_ERROR_NONE;
}

/***************************************************************************
 * Timer                                                                   *
 ***************************************************************************/

Uint32 ENG_Timer(Uint32 interval, void *param) {
	static int intervals = 0;

	/* Count the units of 10ms and set a flag which enables ENG_Woden to proceed */
	if (eng_op_env.cycle_limit) {
		intervals++;
		if (intervals >= 100 / eng_op_env.cycle_limit) {
			eng_next_cycle = TRUE;
			intervals = 0;
		}
	} else {
		intervals = 0;
	}

	return interval;
}

/***************************************************************************
 * Get Joystick Name                                                       *
 ***************************************************************************/
/* This returns a joystick name that isn't null ;) */

/* On entry: joyid is the SDL joystick number
	On exit: returns a pointer to a string that isn't null */

char *ENG_GetJoystickName(int joyid) {
	static char joyname[256];
	
	if (SDL_JoystickName(joyid) == NULL) {
		strcpy(joyname, "(ENG_UNDEFINED)");
	} else {
		strcpy(joyname, SDL_JoystickName(joyid));
	}

	return joyname;
}

/***************************************************************************
 * Stop SDL                                                                *
 ***************************************************************************/

void ENG_StopSDL(void) {
	SDL_Event event;
	int count;
	
	if (SDL_WasInit(SDL_INIT_VIDEO) != 0) {

		SDL_RemoveTimer(eng_timer_id);

		/* Close any opened joysticks */
		for (count = eng_joystick_count - 1; count >= 0; count--) {
			if (eng_joysticks[count] != NULL) SDL_JoystickClose(eng_joysticks[count]);
			eng_joystick_count--;
		}

		/* Free blend screen surface */
		if (eng_blend_screen != NULL) SDL_FreeSurface(eng_blend_screen);

		/* Free display window icon surface */
		if (eng_icon != NULL) SDL_FreeSurface(eng_icon);
		
		/* Flush any events because sometimes Zaurus SDL gets upset */
		while (SDL_PollEvent(&event));

		/* Shutdown SDL */
		SDL_Quit();
		
		/* Nullify all surfaces so we can check if they've been set */
		eng_screen = eng_blend_screen = eng_icon = NULL;
		
		printf("%s: Stopping\n", __func__);
	}
}

