/* ENG is Not a Game (c) Thunor 2008 <thunorsif@hotmail.com>
www.thunor.org.uk

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */


/* Includes */
#include "ENG_engine.h"


/* Defines */
/*
#define DEBUG_ASCII_TEXT
#define DEBUG_FONT_CACHE
*/

#define ENG_MAX_ASCII_FONTS_CACHED 10
#define ENG_ASCII_FONT_START_CHAR ' '
#define ENG_ASCII_FONT_END_CHAR '~'
#define ENG_ASCII_FONT_CHAR_COUNT (ENG_ASCII_FONT_END_CHAR - ENG_ASCII_FONT_START_CHAR + 1)


/* Function prototypes */
Uint32 ENG_GetPixel(SDL_Surface *surface, int x, int y);


/***************************************************************************
 * Initialise Text                                                         *
 ***************************************************************************/

void ENG_InitText(void) {
}

/***************************************************************************
 * Write ASCII Text                                                        *
 ***************************************************************************/
/* This function writes ASCII text to an existing SDL surface. Normally the
   surace will be one declared specifically within an object as
   img_surface=imgBlank{unique_string} and/or
   img_over_surface=imgBlank{unique_string} but it could be any surface.
   
   The properties of the text to write are defined in an ENG_ASCIIText object :-
   
   text.font
   ---------
   An existing SDL_Surface which contains an ASCII bitmap font from space to tilde
   (squiggle) with character width markers in row zero. Note that if there are no
   markers present then the character width becomes fixed.

   text.target
   -----------
   This is the SDL_Surface that will be the recipient of the text.

   text.img_x/y/w/h
   ----------------
   This defines the region of the surface to wrap text within.

   text.bgcolour
   -------------
   This is the colour to wipe the surface region with if not ENG_UNDEFINED.
   
   text.char_spacing
   -----------------
   By default this function will write the next character following the previous one,
   but this behaviour can be modified by passing an adjustment value.
   
   text.line_spacing
   -----------------
   The default is the height of the surface - 1 line for the character width markers
   but this can be modified by passing an adjustment value.
   
   text.tab_spaces
   ---------------
   The number of spaces (+ char_spacing) that a horizontal tab should represent.
   When a tab is encountered, any proceeding character is lined up with the next
   tab stop.
   
   text.fixed_width
   ----------------
   TRUE will ignore any variable width markers within the font.
   
   text.render
   -----------
   FALSE will not wipe the surface region with the bgcolour or render any text but
   will set next_char.
   
   text.text
   ---------
   This is a pointer to the ASCII text to render.
   
   In addition to the ENG_ASCIIText object, page_line_or_word defines how much text
   will be written. demo_01 uses page_line_or_word=ENG_TEXT_WRITE_LINE along with
   text.render=FALSE to calculate a line count for an entire document for the surface
   region width.
   
   Many fonts will have their dimensions cached and if any more fonts are found then
   the one used the least is replaced.
   
   Control characters line feed \n (0x0a) and horizontal tab \t (0x09) are supported.
   
   Invalid characters and unsupported control codes are simply ignored.
   
   On entry: See above
	On exit: On success :-
			 text->next_char will point to the next char or the terminating 0
			 returns the number of lines written which could be zero
			 On error :-			 
			 returns 0 lines written */

int ENG_WriteASCIIText(ENG_ASCIIText *text, int page_line_or_word) {
	static int init = TRUE;
	static int char_dimensions[ENG_MAX_ASCII_FONTS_CACHED][ENG_ASCII_FONT_CHAR_COUNT][2], font_usage[ENG_MAX_ASCII_FONTS_CACHED];
	static SDL_Surface *font[ENG_MAX_ASCII_FONTS_CACHED];
	int count, cx, x12, bgcolour, font_index, found;
	int block_width, rx, ry, line_count = 0, tab_width;
	char *block_start, *ascii_text;
	SDL_Rect srcrect, dstrect;

	/* Intialise some things the once */
	if (init) {
		init = FALSE;
		for (count = 0; count < ENG_MAX_ASCII_FONTS_CACHED; count++) {
			font[count] = NULL;
			font_usage[count] = 0;
		}
	}

	if (eng_started == FALSE) {
		sprintf(eng_last_error, "%s: ENG has not been started yet", __func__);
	} else if (text == NULL || sizeof(*text) != sizeof(ENG_ASCIIText)) {
		sprintf(eng_last_error, "%s: text is invalid", __func__);
	} else if (text->font == NULL) {
		sprintf(eng_last_error, "%s: text->font is invalid", __func__);
	} else if (text->target == NULL) {
		sprintf(eng_last_error, "%s: text->target is invalid", __func__);
	} else {
		/* Are the font dimensions cached? */
		found = FALSE;
		for (font_index = 0; font_index < ENG_MAX_ASCII_FONTS_CACHED; font_index++) {
			if (font[font_index] == text->font) {
				found = TRUE;
				break;
			}
		}
		/* Not cached so is there a spare slot? */
		if (!found) {
			#ifdef DEBUG_FONT_CACHE
				printf("%s: %p not cached\n", __func__, text->font);
			#endif
			for (font_index = 0; font_index < ENG_MAX_ASCII_FONTS_CACHED; font_index++) {
				if (font[font_index] == NULL) {
					found = TRUE;
					break;
				}
			}
			/* No spare slot so find the least used font */
			if (!found) {
				font_index = 0;
				for (count = 0; count < ENG_MAX_ASCII_FONTS_CACHED; count++) {
					if (font_usage[count] < font_usage[font_index]) {
						font_index = count;
					}
					#ifdef DEBUG_FONT_CACHE
						printf("    %p used %i times\n", font[font_index], font_usage[count]);
					#endif
				}
			}
			#ifdef DEBUG_FONT_CACHE
				printf("    Replacing %p in slot %i used %i times with %p\n", font[font_index], font_index, font_usage[font_index], text->font);
			#endif
			/* Add new font details to cache */
			font[font_index] = text->font;
			font_usage[font_index] = 0;
			/* Record the dimensions of each individual character within the font */
			if (SDL_MUSTLOCK(text->font)) SDL_LockSurface(text->font);
			for (count = 0; count < ENG_ASCII_FONT_CHAR_COUNT; count++) {
				/* Set default fixed width dimensions */
				char_dimensions[font_index][count][0] = count * text->font->w / ENG_ASCII_FONT_CHAR_COUNT;
				char_dimensions[font_index][count][1] = (count + 1) * text->font->w / ENG_ASCII_FONT_CHAR_COUNT - 1;
				x12 = 0;
				/* If width markers exist then use those */
				for (cx = char_dimensions[font_index][count][0]; cx <= char_dimensions[font_index][count][1]; cx++) {
					if (ENG_GetPixel(text->font, cx, 0) == 0) {
						char_dimensions[font_index][count][x12++] = cx;
						if (x12 > 1) break;
					}
				}
			}
			if (SDL_MUSTLOCK(text->font)) SDL_UnlockSurface(text->font);
		}
		/* Increment font usage */
		font_usage[font_index] += 1;
		/* Do we need to wipe the surface region? */
		if (text->render && text->bgcolour != ENG_UNDEFINED) {
			bgcolour = SDL_MapRGB(text->target->format, (text->bgcolour >> 16) & 0xff, (text->bgcolour >> 8) & 0xff, text->bgcolour & 0xff);
			dstrect.x = text->img_x; dstrect.y = text->img_y; 
			dstrect.w = text->img_w; dstrect.h = text->img_h; 
			if ((SDL_FillRect(text->target, &dstrect, bgcolour)) < 0) {
				sprintf(eng_last_error, "%s: Cannot fill surface: %s", __func__, SDL_GetError());
				return 0;	/* Quit now */
			}
		}
		/* Calculate the full width of a horizontal tab including char_spacing */
		if (!text->fixed_width) {
			tab_width = (char_dimensions[font_index][' ' - ENG_ASCII_FONT_START_CHAR][1] - char_dimensions[font_index][' ' - ENG_ASCII_FONT_START_CHAR][0] + 1 + text->char_spacing) * text->tab_spaces;
		} else {
			tab_width = (text->font->w / ENG_ASCII_FONT_CHAR_COUNT + text->char_spacing) * text->tab_spaces;
		}
		/* Write the text within the bounds of the surface region */
		block_start = ascii_text = text->text;
		block_width = 0;
		rx = text->img_x;
		ry = text->img_y;
		while (*ascii_text != 0) {
			/* Accumulate a block of text and calculate its width */
			if (*ascii_text >= ENG_ASCII_FONT_START_CHAR && *ascii_text <= ENG_ASCII_FONT_END_CHAR) {
				if (!text->fixed_width) {
					block_width += char_dimensions[font_index][*ascii_text - ENG_ASCII_FONT_START_CHAR][1] - char_dimensions[font_index][*ascii_text - ENG_ASCII_FONT_START_CHAR][0] + 1;
				} else {
					block_width += text->font->w / ENG_ASCII_FONT_CHAR_COUNT;
				}
				block_width += text->char_spacing;
			} else if (*ascii_text == '\t') {
				block_width += tab_width - (rx - text->img_x + block_width) % tab_width;
			}
			/* These things trigger writing :-
					Spaces
					Line feeds 0xa
					Hyphens not followed by another hyphen or space
					End of string 0x0 */
			if (*ascii_text == ' ' || *ascii_text == '\n' || (*ascii_text == '-' && *(ascii_text + 1) != '-' && *(ascii_text + 1) != ' ') || *(ascii_text + 1) == 0) {
				/* Fit the text within the surface region */
				if (rx + block_width > text->img_x + text->img_w && rx != text->img_x) {
					rx = text->img_x;
					ry += text->font->h - 1 + text->line_spacing;
					line_count++;
				}
				/* Only write as much as is requested */
				if (page_line_or_word == ENG_TEXT_WRITE_LINE) {
					if (ry != text->img_y) break;
				} else if (page_line_or_word == ENG_TEXT_WRITE_PAGE) {
					if (ry + text->font->h - 1 + text->line_spacing > text->img_y + text->img_h) break;
				}
				/* Write block */
				while (block_start <= ascii_text) {
					if (*block_start == '\n') {
						rx = text->img_x;
						ry += text->font->h - 1 + text->line_spacing;
						line_count++;
					} else if (*block_start == '\t') {
						rx += tab_width - (rx - text->img_x) % tab_width;
					} else if (*block_start >= ENG_ASCII_FONT_START_CHAR && *block_start <= ENG_ASCII_FONT_END_CHAR) {
						if (!text->fixed_width) {
							srcrect.x = char_dimensions[font_index][*block_start - ENG_ASCII_FONT_START_CHAR][0];
							srcrect.w = char_dimensions[font_index][*block_start - ENG_ASCII_FONT_START_CHAR][1] - char_dimensions[font_index][*block_start - ENG_ASCII_FONT_START_CHAR][0] + 1;
						} else {
							srcrect.x = text->font->w / ENG_ASCII_FONT_CHAR_COUNT * (*block_start - ENG_ASCII_FONT_START_CHAR);
							srcrect.w = text->font->w / ENG_ASCII_FONT_CHAR_COUNT;
						}
						if (text->render) {
							srcrect.y = 1;
							srcrect.h = text->font->h - 1;
							dstrect.x = rx;
							dstrect.y = ry;
							/* Blit surface */
							if ((SDL_BlitSurface(text->font, &srcrect, text->target, &dstrect)) < 0) {
								sprintf(eng_last_error, "%s: Cannot blit surface: %s", __func__, SDL_GetError());
								return 0;	/* Quit now */
							}
						}
						rx += srcrect.w + text->char_spacing;
					}
					block_start++;
				}
				/* Only write as much as is requested */
				if (page_line_or_word == ENG_TEXT_WRITE_WORD) break;
				block_width = 0;
			}
			ascii_text++;
		}
		#ifdef DEBUG_ASCII_TEXT
			printf("%s: Next char: %c (0x%x)\n", __func__, *block_start, *block_start);
		#endif
		text->next_char = block_start;
	}

	return line_count;
}


/***************************************************************************
 * Get Pixel                                                               *
 ***************************************************************************/
/* This function is based upon getpixel() from the SDL documentation.
   If surfaces must be locked/unlocked then the caller is responsible for this */

/*  On exit: returns a value which can be converted to RGB using SDL_GetRGB() */

Uint32 ENG_GetPixel(SDL_Surface *surface, int x, int y) {
	Uint32 retpixel;
	int bpp;
	Uint8 *offset;

	bpp = surface->format->BytesPerPixel;
	offset = (Uint8*)surface->pixels + y * surface->pitch + x * bpp;

	if (bpp == 1) {
		retpixel = *offset;
	} else if (bpp == 2) {
		retpixel = *(Uint16*)offset;
	} else if (bpp == 3) {
		if (SDL_BYTEORDER == SDL_BIG_ENDIAN) {
			retpixel = offset[0] << 16 | offset[1] << 8 | offset[2];
		} else {
			retpixel = offset[0] | offset[1] << 8 | offset[2] << 16;
		}
	} else if (bpp == 4) {
		retpixel = *(Uint32*)offset;
	} else {
		retpixel = 0;
	}
	
	return retpixel;
}







